import os
from collections import defaultdict
from typing import Callable, DefaultDict, Hashable, NoReturn, TypedDict, cast

from django import template
from django.template.base import Node, NodeList, Parser, Token
from django.template.context import Context
from django.template.library import parse_bits
from django.template.loader import get_template

from django_slots.templatetags.slot_tags import SlotNode
from django_slots.utils import pascalcase_to_snakecase

DEFAULT_SLOT_NAME = "slot"


SlotsDict = DefaultDict[str, str]


class ContextData(TypedDict):
    slot: str
    slots: SlotsDict


class ComponentError(Exception):
    pass


class ComponentValidationError(Exception):
    pass


class Component:
    inline_tag_name = "{name}/"
    block_tag_names = "{name}", "/{name}"
    namespace = ""
    name = None
    template_name = ""

    @classmethod
    def get_name(cls) -> str:
        if cls.name:
            return cls.name
        return pascalcase_to_snakecase(cls.__name__)

    @classmethod
    def get_namespaced_name(cls) -> str:
        name = cls.get_name()
        if cls.namespace:
            return f"{cls.namespace}:{name}"
        return name

    def get_template_name(self) -> str:
        if self.template_name:
            return self.template_name
        return os.path.join("components", self.namespace, f"{self.get_name()}.html")

    @classmethod
    def get_inline_tag_name(cls) -> str:
        return cls.inline_tag_name.format(name=cls.get_namespaced_name())

    @classmethod
    def get_block_tag_names(cls) -> tuple[str, str]:
        name = cls.get_namespaced_name()
        start_tag, end_tag = cls.block_tag_names
        return start_tag.format(name=name), end_tag.format(name=name)

    @classmethod
    def validation_error(cls, msg: str) -> NoReturn:
        raise ComponentValidationError(f"{cls.get_name()} component {msg}")

    @classmethod
    def inline_compile_function(
        cls,
    ) -> Callable[[Parser, Token], "ComponentNode"]:
        def compile_func(parser: Parser, token: Token):
            component_params, component_kwargs = parse_bits(
                parser=parser,
                bits=token.split_contents()[1:],
                params=[],
                takes_context=False,
                name=cls.get_inline_tag_name(),
                varargs=None,
                varkw=[],
                defaults=None,
                kwonly=[],
                kwonly_defaults=None,
            )
            component = cls()
            return ComponentNode(component=component, kwargs=component_kwargs, slots=[])

        return compile_func

    @classmethod
    def block_compile_function(
        cls,
    ) -> Callable[[Parser, Token], "ComponentNode"]:
        def compile_func(parser: Parser, token: Token):
            nodelist = parser.parse(parse_until=[cls.get_block_tag_names()[1]])
            parser.delete_first_token()
            component_params, component_kwargs = parse_bits(
                parser=parser,
                bits=token.split_contents()[1:],
                params=[],
                takes_context=False,
                name=cls.get_block_tag_names()[1],
                varargs=None,
                varkw=[],
                defaults=None,
                kwonly=[],
                kwonly_defaults=None,
            )
            component = cls()
            slots = component._find_slot_nodes(nodelist=nodelist)
            return ComponentNode(
                component=component, kwargs=component_kwargs, slots=slots
            )

        return compile_func

    @staticmethod
    def _find_slot_nodes(*, nodelist: NodeList) -> list[SlotNode]:
        slot_nodes = nodelist.get_nodes_by_type(SlotNode)
        remaining_nodes = NodeList(node for node in nodelist if node not in slot_nodes)
        return [SlotNode(name=DEFAULT_SLOT_NAME, nodelist=remaining_nodes), *slot_nodes]

    def get_context_data(
        self, slots: SlotsDict, **kwargs: dict[Hashable, object]
    ) -> ContextData:
        return cast(
            ContextData,
            {"slot": slots[DEFAULT_SLOT_NAME], "slots": slots, **kwargs},
        )


class ComponentNode(Node):
    def __init__(
        self, *, component: Component, kwargs: dict, slots: list[SlotNode]
    ) -> None:
        self.component = component
        self.kwargs = kwargs
        self.slots = slots

    def render(self, context: Context) -> str:
        # Convert to defaultdict to prevent KeyError with use of |default template filter
        slots: SlotsDict = defaultdict(
            str, {slot.name: slot.render(context) for slot in self.slots}
        )
        resolved_kwargs = {
            key: value.resolve(context) for key, value in self.kwargs.items()
        }
        try:
            context = self.component.get_context_data(slots, **resolved_kwargs)
        except TypeError as error:
            # Reraise unexpected arguments and required arguments errors
            # annotated with component name.
            raise self.component.validation_error(str(error)) from error
        return get_template(self.component.get_template_name()).render(context)


class Library(template.Library):
    def component(self, component_class: type[Component]) -> type[Component]:
        """
        Register inline and block component
        """
        self.inline_component(component_class)
        self.block_component(component_class)
        return component_class

    def inline_component(self, component_class: type[Component]) -> type[Component]:
        """
        Register an inline component
        """
        self.tag(
            component_class.get_inline_tag_name(),
            component_class.inline_compile_function(),
        )
        return component_class

    def block_component(self, component_class: type[Component]) -> type[Component]:
        """
        Register a block component
        """
        self.tag(
            component_class.get_block_tag_names()[0],
            component_class.block_compile_function(),
        )
        return component_class



import re


def pascalcase_to_snakecase(string: str) -> str:
    """
    Convert string from PascalCase to snake_case
    """
    return (
        re.sub("(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))", "_\\1", string)
        .lower()
        .strip("_")
    )



from django.template import Library, TemplateSyntaxError
from django.template.base import Node, NodeList, Parser, Token
from django.template.context import Context

register = Library()


@register.tag(name="slot")
def slot(parser: Parser, token: Token) -> "SlotNode":
    bits = token.contents.split()
    if len(bits) > 2:
        raise TemplateSyntaxError(f"'slot' tag got unexpected arguments {bits[2:]!r}")
    from django_slots.components import DEFAULT_SLOT_NAME

    slot_name = DEFAULT_SLOT_NAME
    if len(bits) == 2:
        slot_name = bits[1]

    nodelist = parser.parse(parse_until=["/slot"])
    parser.delete_first_token()

    return SlotNode(name=slot_name, nodelist=nodelist)


class SlotNode(Node):
    def __init__(self, *, name: str, nodelist: NodeList) -> None:
        self.name = name
        self.nodelist = nodelist

    def render(self, context: Context) -> str:
        return self.nodelist.render(context)

    def __repr__(self):
        return f"<SlotNode name={self.name}>"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_slots', 'django_slots.templatetags']

package_data = \
{'': ['*']}

install_requires = \
['Django>=4.0.2,<5.0.0']

setup_kwargs = {
    'name': 'django-slots',
    'version': '0.2.6',
    'description': 'django_slots = inclusion tag + blocks',
    'long_description': '# django_slots\n\nAllows multiline strings to be captured and passed to template tags.\n\n## Demo\n\n1. Register a component\n  \n```python\n# app/templatetags/component_tags.py\nfrom django_slots import Library, Component\n\nregister = Library()\n\n\n@register.block_component\nclass Details(Component):\n  pass\n```\n\n2. Create a template\n\n```html+django\n{# app/templates/components/details.html #}\n<details>\n<summary>{{ summary|default:slots.summary }}</summary>\n{{ slot }}\n</details>\n```\n\nUsage:\n\n```html+django\n{% load component_tags %}\n{% load slot_tags %}\n\n{% details summary="the summary" %}\n  the default slot\n{% /details %}\n\n{% details %}\n  {% slot summary %}the <b>summary</b>{% /slot %}\n  the default slot\n{% /details %}\n```\n\n## Installation\n\n```shell\npip install django-slots\n```\n\n```python\nINSTALLED_APPS = [\n    # ...\n    \n    \'django_slots\',\n]\n```\n\n## Slots\n\nUse `{% slot name %}{% /slot %}` to capture and name a slot. These slots will be available in the template in a dictionary called `slots`. eg `{{ slots.name }}`.\n\nAny lines not surrounded by a slot tag will be available in thh component template as `{{ slot }}`.\n\n## Custom name\n\nBy default, the template tag name will be the Component class name converted to snake case. Use the `name` attribute to override.\n\neg:\n\n```python\nfrom django_slots import Component, Library\n\nregister = Library()\n\n\n@register.inline_component\nclass Button(Component):\n    name = \'btn\'\n```\n\n```html+django\n{% btn %}\n```\n\n## Change tag name\n\nBy default, inline tags are named `"{name}/"` and block tags are named `"{name}", "/{name}"`. To change this format specify `inline_tag_name` and `block_tag_names` attributes.\n\neg:\n\n```python\nfrom django_slots import Component, Library\n\nregister = Library()\n\n\nclass AppComponent(Component):\n    inline_tag_name = "{name}end"\n    block_tag_names = "{name}", "end{name}"    \n\n    \n@register.component\nclass Button(AppComponent):\n    pass\n```\n\n```html+django\n{% buttonend %}\n\n{% button %}{% endbutton %}\n```\n\n## Inline only template tag\n\nUse `@register.inline_component` to only allow `{% inline/ %}` use.\n\n## Block only template tag\n\nUse `@register.block_component` to only allow `{% block %}{% /block %}` use.\n\n## Validate arguments\n\nImplement `def get_content_data(slots, **kwargs)` to validate arguments. \n\neg:\n\n```python\nfrom django_slots import Component, Library\n\nregister = Library()\n\n\n@register.component\nclass Button(Component):\n    STYLE = ["primary", "secondary"]\n    def get_context_data(self, slots, *, style: str):\n        if style not in self.STYLE:\n            raise self.validation_error(f"style {style!r} not in {self.STYLE!r}")\n        return super().get_context_data(slots, style=style)\n```\n\n## Namespace components\n\nComponents can be namespaced which is useful for creating a third party app.\n\n```python\nfrom django_slots import Library, Component\n\nregister = Library()\n\n\nclass NHSUKComponent(Component):\n    namespace = \'nhsuk\'\n\n\n@register.component\nclass Button(NHSUKComponent):\n    pass\n```\n\n```html+django\n{% nhsuk:button %}\n  Save and continue\n{% /nhsuk:button %}\n```\n\nSee https://github.com/nwjlyons/nhsuk-components\n',
    'author': 'Neil Lyons',
    'author_email': 'nwjlyons@googlemail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nwjlyons/django_slots',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
